/* IRC ProxyScan, m_cache.c
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * To keep in memory last recent scanned clients' IPs.
 *
 * $Id: m_cache.c,v 1.14 2001/11/21 15:43:23 mbuna Exp $
 */

#undef CACHE_TEST   /* for debugging and testing the cache */


#ifdef CACHE_TEST

#define CACHESIZE   1000000   /* size of the cache to test */
#define CACHEINJECT 3000000   /* simulate an injection of IPs in the cache:
                               * number of ips to inject */
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>          /* time_t */
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <sys/types.h>
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifndef CACHE_TEST

#include "m_structs.h"
#include "typedefs.h"

#include "conf.h"
#include "debug.h"

#endif /* ifndef CACHE_TEST */


#include "m_cache.h"
#include "cachehash.h"

/* Should be defined in sys/stat.h */
#ifndef S_IREAD
#define S_IREAD 0000400
#endif
#ifndef S_IWRITE
#define S_IWRITE 0000200
#endif

#ifdef CACHE_TEST
#define xmalloc malloc
#define xfree free
#else
#include "xmalloc.h"
#include "m_log.h"
#endif /* CACHE_TEST */

#ifndef CACHE_TEST
/* To avoid threads conflicts when we're cache_dump'ing, we use here the
 * same mutex as the one used for protecting array_fetchitem.
 * That's cool because cache_insert and cache_fetch are only called from
 * the array_fetchitem() function.
 */
extern pthread_mutex_t mutex_scanext;
#endif

cachehtab *cachetab = NULL;
/* PRIVATE */
static cacheItem* cachebuf = NULL;
static unsigned long cachepos;

#ifdef CACHE_TEST
int
main(void)
{
   unsigned int u, found=0;
   struct in_addr addr;
   time_t now;
   time_t dummy;
   
   time(&now);
    
   cache_init();
   
   printf("main: injecting %u IPs\n", CACHEINJECT);
   
   for (u = 0; u < CACHEINJECT; u++) {
      addr.s_addr = u;
      cache_insert(&addr, now);
   }
   
   printf("Insert time = %ld\n", time(NULL) - now);
   time(&now);
   
   printf("cache_status = %u\n", cache_status());
   
   printf("cache_fetch: last: %u\n", cache_fetch(&addr));
   
   addr.s_addr = u - 3000;
   printf("cache_fetch: first: %u\n", cache_fetch(&addr));
   
   for (u = 0; u < CACHEINJECT; u++) {
      addr.s_addr = u;
      dummy = cache_fetch(&addr);
      if (dummy) found++;
   }
   printf("main: retrieved %u IPs\n", found);
   printf("Fetch time = %u\n", time(NULL) - now);
   
   return 0;
}
#endif /* ifdef CACHE_TEST */

void
cache_init()
{
   double logsize;
   
#ifdef CACHE_TEST
   size_t allocsize = CACHESIZE * sizeof(cacheItem); /* simulation */
#else
   size_t allocsize = MyPerform->cachesize * sizeof(cacheItem);
#endif

#ifdef CACHE_TEST
   printf("cache_init: allocsize=%d (%dK)\n", (int)allocsize, (int)allocsize/1024);
#else
   Debug((DBLMAX, "cache_init: allocsize=%d (%dK)\n", (int)allocsize, (int)allocsize/1024));
#endif
  
   cachebuf = xmalloc(allocsize);
   memset(cachebuf, 0, allocsize);
   
#ifdef CACHE_TEST  
   logsize = log((double)CACHESIZE)/log((double)2);
#else
   logsize = log((double)MyPerform->cachesize)/log((double)2);
#endif
   
#ifdef CACHE_TEST
   printf("cache_init: logsize = %g (%d)\n", logsize, (int)logsize + 1);
#else
   Debug((DBLMAX, "cache_init: logsize = %g (%d)\n", logsize, (int)logsize + 1));
#endif
   cachetab = cache_hcreate((int)logsize + 1); /* ln(CACHESIZE)/ln(2) */

#ifdef CACHE_TEST
   printf("cache_init: alloc = %luK\n", sizeof(cachehitem *)*((ub4)1<<cachetab->logsize)/1024);
#else
   Debug((DBLMAX, "cache_init: alloc = %luK\n", sizeof(cachehitem *)*((ub4)1<<cachetab->logsize)/1024));
#endif
     
   cachepos = 0;
}

void
cache_free()
{
   if (cachebuf) {
      xfree(cachebuf);
      cachebuf = NULL;
   }
   if (cachetab)
      cache_hdestroy(cachetab);
}


void
cache_insert(struct in_addr *addr, time_t last)
{
   cacheItem *citem = &cachebuf[cachepos];
   ub4 x; /* hash value for new item */
   
   if (cache_hfindx(cachetab, (char*)addr, &x))
      cache_hstuff(cachetab) = last;
   else {
      
      /* is cache buffer full ? */
#ifdef CACHE_TEST
      if(cache_hcount(cachetab) < CACHESIZE) 
#else
      if(cache_hcount(cachetab) < MyPerform->cachesize)
#endif
      {
         memcpy(&citem->addr, addr, sizeof(struct in_addr));
         
         /* we checked hfind, now use my cache_haddx (which dont check for dup) */
         cache_haddx(cachetab, (char*)&citem->addr, last, x);
      
      } else {
         
         /* cache buffer is full: replace the entry */
         if (cache_hfind(cachetab, (char*)&citem->addr)) {
            cache_hdel(cachetab);
            memcpy(&citem->addr, addr, sizeof(struct in_addr));
            cache_haddx(cachetab, (char*)&citem->addr, last, x);
         }
      }
   }
   /* Update position in the array */
#ifdef CACHE_TEST
   cachepos = (cachepos + 1 < CACHESIZE) ? cachepos + 1 : 0 ;
#else
   cachepos = (cachepos + 1 < MyPerform->cachesize) ? cachepos + 1 : 0 ;
#endif
}


/*
 *  0.9      12.65     0.12  3000002     0.00     0.00  cache_fetch [4]  :)
 */
time_t
cache_fetch(struct in_addr *addr)
{
   if(cache_hfind(cachetab, (char*)addr))
      return cache_hstuff(cachetab);
      
   return 0;
}

void
cache_remove(struct in_addr *addr)
{
   if (cache_hfind(cachetab, (char*)addr))
      cache_hstuff(cachetab) = 0;
}

size_t
cache_getsize()
{
#ifdef CACHE_TEST
   return CACHESIZE;
#else
   return MyPerform->cachesize;
#endif
}

size_t
cache_status()
{
   return ((size_t)cache_hcount(cachetab));
}

/**************************** CACHE DUMP TO DISK ****************************/

/* Call that after cache_init, it loads the cache records from disk
 * (data/cache)
 */
int
cache_load()
{
   int fd, ret = 0;
   size_t alloc;
   void *readbuf, *p;
   struct in_addr ip;
   time_t t;
   ssize_t len;
   
   fd = open("data/cache", O_RDONLY, 0);
   
   if (fd < 0)
      return 0;
   
   alloc = 1<<14;
   
   readbuf = xmalloc(alloc); /* Read blocks of 16K */
   
   while ((len = read(fd, readbuf, alloc))
      && (len >= (sizeof(struct in_addr) + sizeof(time_t)))) {
      p = readbuf;
      do {
         memcpy(&ip, p, sizeof(struct in_addr));
         memcpy(&t, p + sizeof(struct in_addr), sizeof(time_t));
         cache_insert(&ip, t); ret++;
         p += (sizeof(struct in_addr) + sizeof(time_t));
      } while ((len -= (sizeof(struct in_addr) + sizeof(time_t))) > 0);
   }
   
   close(fd);
   
   xfree(readbuf);
   
   return ret;
}

void
cache_dump(int protect)
{
   int fd;
   long len = 0;
   void *writebuf, *p;
   size_t alloc;
   time_t now;

#ifndef CACHE_TEST
   if (protect)
      pthread_mutex_lock(&mutex_scanext);
#endif
   
   fd = open("data/cache", O_WRONLY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
   if (fd < 0)
      return;
   
   alloc = 1<<14;
   
   writebuf = p = xmalloc(alloc); /* Write blocks of 16K */
   
   time(&now);
   
   if (cache_hfirst(cachetab)) do {
      
      /* Do not cache IPs which may have not been scanned yet. */
      if (now - cache_hstuff(cachetab) >= 60) {
         memcpy(p, cache_hkey(cachetab), sizeof(struct in_addr));
         p += sizeof(struct in_addr);
         memcpy(p, &cache_hstuff(cachetab), sizeof(time_t));
         p += sizeof(time_t);
         
         len += (sizeof(struct in_addr) + sizeof(time_t));
         
         /* See if we're going to overflow next time... */
         if (len + (sizeof(struct in_addr) + sizeof(time_t)) > alloc) {
            if (write(fd, writebuf, len) < 0) {
               log_error("m_cache:cache_dump: write < 0 (%s)", strerror(errno));
               break;
            }
            len = 0;
            p = writebuf;
         }
      }
      
   } while (cache_hnext(cachetab));
   
   if (len > 0)
      write(fd, writebuf, len);
   
   close(fd);
   
   xfree(writebuf);
   
#ifndef CACHE_TEST
   if (protect)
      pthread_mutex_unlock(&mutex_scanext);
#endif
}

/* m_cache.c */
